/* Modification of RT-FIFO devices (copyright Michael Barabanov, 1997)      */
/* by Paolo Mantegazza, for their use with RTAI module. They work in the    */
/* same way of their original counterparts but now you can chose either a   */
/* task queue for bottom halves or, an immediate wake up plus assigning a   */
/* static user buffer.                                                      */
/* See functions: rtf_create_using_bh and rtf_create_using_bh_and_usr_buf.  */

// A (very) small grain locking, i.e. a lock per fifo + another for sysrqs,
// is used for efficiency. However a global locking is also possibile. (It was 
// used during development mainly to test the global_... stuff). If you'd like
// to try it uncomment the line below.
//#define USE_GLOBAL_LOCK


#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/poll.h>
#include <linux/tqueue.h>
#include <linux/interrupt.h>
#include <linux/major.h>

#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/io.h>
#include <linux/smp.h>

#include <rtai.h>
#include <rtai_fifos.h>


#define RTF_NO 64

struct rt_fifo_struct {
	int open;
	int user_open;
	char *base;
	int bufsize;
	int start;
	int len;
	int (*user_handler) (unsigned int fifo);
	struct wait_queue *wait;
	struct tq_struct wake_up_task;
	spinlock_t spinlock;
	struct task_struct *sleeps;
	task_queue *bh_list;
};

static struct rt_fifo_struct fifoes[RTF_NO];
static struct {int srq, in, out, minors[RTF_NO]; spinlock_t spinlock;} sysrq;

#ifdef USE_GLOBAL_LOCK
#define rt_fifo_critical(flags) 	flags = rt_global_save_flags_and_cli()
#define rt_end_fifo_critical(flags) 	rt_global_restore_flags(flags) 
#define rt_sysrq_critical(flags)  	flags = rt_global_save_flags_and_cli()
#define rt_end_sysrq_critical(flags)	rt_global_restore_flags(flags) 
#else
#define rt_fifo_critical(flags) \
{ flags = rt_spin_lock_irqsave(&RTF_SPINLOCK(minor)); }
#define rt_end_fifo_critical(flags) \
{ rt_spin_unlock_irqrestore(flags, &RTF_SPINLOCK(minor)); }
#define rt_sysrq_critical(flags) \
{ flags = rt_spin_lock_irqsave(&sysrq.spinlock); }
#define rt_end_sysrq_critical(flags) \
{ rt_spin_unlock_irqrestore(flags, &sysrq.spinlock); }
#endif

#define RTF_ADDR(minor)		(&fifoes[minor])

#define RTF_OPEN(minor)		(RTF_ADDR(minor)->open)
#define RTF_USER_OPEN(minor)	(RTF_ADDR(minor)->user_open)
#define RTF_BASE(minor)		(RTF_ADDR(minor)->base)
#define RTF_BUF(minor)		(RTF_ADDR(minor)->bufsize)
#define RTF_START(minor)	(RTF_ADDR(minor)->start)
#define RTF_HANDLER(minor)	(RTF_ADDR(minor)->user_handler)
#define RTF_LEN(minor)		(RTF_ADDR(minor)->len)
#define RTF_FREE(minor)		(RTF_BUF(minor) - RTF_LEN(minor))
#define RTF_WAIT(minor)		(RTF_ADDR(minor)->wait)
#define RTF_SLEEPS(minor)	(RTF_ADDR(minor)->sleeps)
#define RTF_WAKE_UP_TASK(minor)	(RTF_ADDR(minor)->wake_up_task)
#define RTF_BH_LIST(minor)	(RTF_ADDR(minor)->bh_list)
#define RTF_SPINLOCK(minor)	(RTF_ADDR(minor)->spinlock)

#define RTF_WRAP(minor,pos)	((pos) < RTF_BUF(minor) ? (pos) : (pos) - RTF_BUF(minor))
#define RTF_END(minor)		RTF_WRAP(minor, RTF_START(minor) + RTF_LEN(minor))
#define RTF_EMPTY(minor)	(RTF_LEN(minor) == 0)
#define RTF_FULL(minor)		(RTF_FREE(minor) == 0)
#define RTF_MAX_RCHUNK(minor)	(RTF_BUF(minor) - RTF_START(minor))
#define RTF_MAX_WCHUNK(minor)	(RTF_BUF(minor) - RTF_END(minor))


static inline int check_blocked(sigset_t signal, sigset_t blocked)
{
	int i = 0;
	do {
		if ( signal.sig[i] & ~blocked.sig[i] ) {
			return 1;
		}
	} while (++i < _NSIG_WORDS);
	return 0;
}

static inline void rt_fifo_queue_task(unsigned int minor)
{
	unsigned long interrupt_state;
	rt_sysrq_critical(interrupt_state);
	if (RTF_SLEEPS(minor) && RTF_SLEEPS(minor) == RTF_WAIT(minor)->task &&
			RTF_SLEEPS(minor)->state == TASK_INTERRUPTIBLE) { 
		RTF_SLEEPS(minor) = 0;
		sysrq.minors[sysrq.in] = minor;
		sysrq.in = (sysrq.in + 1) & (RTF_NO-1);
		rt_pend_linux_srq(sysrq.srq);
	}
	rt_end_sysrq_critical(interrupt_state);
}

void rt_fifo_sysrq_handler(void)
{
	unsigned int minor;
	unsigned long interrupt_state;
	rt_sysrq_critical(interrupt_state);
	while (sysrq.out != sysrq.in) {
		minor = sysrq.minors[sysrq.out];
		sysrq.out = (sysrq.out + 1) & (RTF_NO-1);
		rt_end_sysrq_critical(interrupt_state);
		if (RTF_BH_LIST(minor)) {
			queue_task(&RTF_WAKE_UP_TASK(minor), RTF_BH_LIST(minor));
			if (RTF_BH_LIST(minor) == &tq_immediate) {
				mark_bh(IMMEDIATE_BH);
			}
		} else {
			wake_up_interruptible(&RTF_WAIT(minor));
		}
		rt_sysrq_critical(interrupt_state);
	}
	rt_end_sysrq_critical(interrupt_state);
}


static void rtf_wake_up(void *p)
{
	struct rt_fifo_struct *fifo_ptr = (struct rt_fifo_struct *) p;
	wake_up_interruptible(&(fifo_ptr->wait));
}

void rtf_reset(unsigned int minor)
{
	unsigned long interrupt_state;
	rt_fifo_critical(interrupt_state);
	RTF_START(minor) = 0;
	RTF_LEN(minor) = 0;
	rt_end_fifo_critical(interrupt_state);
}

/* These are for use in the init and exit code of real-time modules
   DO NOT call these from a RT task  */

int rtf_resize(unsigned int minor, int size)
{
	void *mem;
	void *old;
	unsigned long interrupt_state;

	if (minor >= RTF_NO) {
		return -ENODEV;
	}
	mem = kmalloc(size, GFP_KERNEL);
	if (!mem) {
		return -ENOMEM;
	}
	old = RTF_BASE(minor);


	rt_fifo_critical(interrupt_state);
	RTF_BASE(minor) = mem;
	RTF_BUF(minor) = size;
	RTF_START(minor) = 0;
	RTF_LEN(minor) = 0;
	rt_end_fifo_critical(interrupt_state);

	if (RTF_OPEN(minor)) {
		kfree(old);
	}
	return size;
}

// bh_list can be: &tq_scheduler, &tq_timer, &tq_immediate (other tqs never
// tested). If 0 wake_up_interruptible is called directly without resorting to
// bottom halves, the old plain rtf_create without a bh_list defaults to 
// &tq_immediate (see the function below).
int rtf_create_using_bh(unsigned int minor, int size, task_queue *bh_list)
{
	int result;

	if (minor >= RTF_NO) {
		return -ENODEV;
	}
	if (RTF_OPEN(minor)) {
		return -EBUSY;
	}
	if ((result = rtf_resize(minor, size)) < 0) {
		printk("RT_FIFO: rtf_create: can't allocate memory.\n");
		return result;
	}
	RTF_OPEN(minor) = 1;
	RTF_SLEEPS(minor) = 0;
	RTF_USER_OPEN(minor) = 0;
	RTF_HANDLER(minor) = 0;

	RTF_WAKE_UP_TASK(minor).next = 0;
	RTF_WAKE_UP_TASK(minor).sync = 0;
	RTF_WAKE_UP_TASK(minor).data = RTF_ADDR(minor);
	RTF_WAKE_UP_TASK(minor).routine = &rtf_wake_up;

	RTF_WAIT(minor) = 0;

      	spin_lock_init(&RTF_SPINLOCK(minor));
	RTF_BH_LIST(minor) = bh_list;

	return 0;
}


int rtf_create(unsigned int minor, int size)
{
	return rtf_create_using_bh(minor, size, &tq_immediate);
}


int rtf_create_using_bh_and_usr_buf(unsigned int minor, char *buf, int size, task_queue *bh_list)
{
	if (minor >= RTF_NO) {
		return -ENODEV;
	}
	if (RTF_OPEN(minor)) {
		return -EBUSY;
	}

	RTF_BASE(minor) = buf;
	RTF_BUF(minor) = size;
	RTF_START(minor) = 0;
	RTF_LEN(minor) = 0;

	RTF_OPEN(minor) = 1;
	RTF_SLEEPS(minor) = 0;
	RTF_USER_OPEN(minor) = 0;
	RTF_HANDLER(minor) = 0;

	RTF_WAKE_UP_TASK(minor).next = 0;
	RTF_WAKE_UP_TASK(minor).sync = 0;
	RTF_WAKE_UP_TASK(minor).data = RTF_ADDR(minor);
	RTF_WAKE_UP_TASK(minor).routine = &rtf_wake_up;

	RTF_WAIT(minor) = 0;

      	spin_lock_init(&RTF_SPINLOCK(minor));
	RTF_BH_LIST(minor) = bh_list;

	return 0;
}


int rtf_destroy(unsigned int minor)
{
	if (minor >= RTF_NO) {
		return -ENODEV;
	}
	if (RTF_USER_OPEN(minor)) {
		return -EINVAL;
	}
	if (!RTF_OPEN(minor)) {
		return -EINVAL;
	}
	RTF_OPEN(minor) = 0;
	kfree(RTF_BASE(minor));
	return 0;
}


int rtf_destroy_using_usr_buf(unsigned int minor)
{
	if (minor >= RTF_NO) {
		return -ENODEV;
	}
	if (RTF_USER_OPEN(minor)) {
		return -EINVAL;
	}
	if (!RTF_OPEN(minor)) {
		return -EINVAL;
	}
	RTF_OPEN(minor) = 0;
	return 0;
}


int rtf_create_handler(unsigned int minor, int (*handler) (unsigned int fifo))
{
	if (minor >= RTF_NO || !RTF_OPEN(minor) || !handler) {
		return -EINVAL;
	}
	RTF_HANDLER(minor) = handler;
	return 0;
}


/* these can be called from RT tasks and interrupt handlers */

int rtf_put(unsigned int minor, void *buf, int count)
{
	unsigned long interrupt_state;
	int chars = 0, free = 0, written = 0;
	char *pipebuf;

	if (minor >= RTF_NO) {
		return -ENODEV;
	}
	if (!RTF_OPEN(minor))
		return -EINVAL;

	rt_fifo_critical(interrupt_state);
	if (RTF_FREE(minor) < count) {
		rt_end_fifo_critical(interrupt_state);
		rt_fifo_queue_task(minor);
		return -ENOSPC;
	}
	while (count > 0 && (free = RTF_FREE(minor))) {
		chars = RTF_MAX_WCHUNK(minor);
		if (chars > count)
			chars = count;
		if (chars > free)
			chars = free;
		pipebuf = RTF_BASE(minor) + RTF_END(minor);
		written += chars;
		RTF_LEN(minor) += chars;
		count -= chars;
		memcpy(pipebuf, buf, chars);
		buf += chars;
	}
	rt_end_fifo_critical(interrupt_state);
	rt_fifo_queue_task(minor);
	return count;
}


int rtf_get(unsigned int minor, void *buf, int count)
{
	unsigned long interrupt_state;
	int chars = 0, size = 0, read = 0;
	char *pipebuf;

	if (minor >= RTF_NO) {
		return -ENODEV;
	}
	if (!RTF_OPEN(minor))
		return -EINVAL;

	rt_fifo_critical(interrupt_state);
	while (count > 0 && (size = RTF_LEN(minor))) {
		chars = RTF_MAX_RCHUNK(minor);
		if (chars > count)
			chars = count;
		if (chars > size)
			chars = size;

		read += chars;
		pipebuf = RTF_BASE(minor) + RTF_START(minor);
		RTF_START(minor) += chars;
		RTF_START(minor) = RTF_WRAP(minor, RTF_START(minor));
		RTF_LEN(minor) -= chars;
		count -= chars;
		memcpy(buf, pipebuf, chars);
		buf += chars;
	}
	rt_end_fifo_critical(interrupt_state);
	rt_fifo_queue_task(minor);
	return read;
}


/* 
 * these are file_operations functions
 * called by user tasks via the fops structure 
 */

static int rtf_open(struct inode *inode, struct file *filp)
{
	unsigned int minor = MINOR(inode->i_rdev);

	if (minor >= RTF_NO)
		return -ENODEV;

	if (!RTF_OPEN(minor)) {
		return -EINVAL;
	}
	if (RTF_USER_OPEN(minor)) {
		return -EBUSY;
	}
	RTF_USER_OPEN(minor) = 1;

	MOD_INC_USE_COUNT;
	return 0;
}


static int rtf_release(struct inode *inode, struct file *file)
{
	unsigned int minor = MINOR(inode->i_rdev);

	if (!RTF_USER_OPEN(minor)) {
		printk("RT-FIFO: release of unopened descriptor inode = %d.\n",
							(int)inode->i_ino);
		return 0;
	}
	RTF_SLEEPS(minor) = 0;
	RTF_USER_OPEN(minor) = 0;
	MOD_DEC_USE_COUNT;
	return 0;
}


static loff_t rtf_llseek(struct file *file, loff_t offset, int origin)
{
	return -ESPIPE;
}


static ssize_t rtf_read(struct file *filp, char *buf, size_t count, loff_t* ppos)
{
	struct inode* inode = filp->f_dentry->d_inode;
	unsigned int minor = MINOR(inode->i_rdev);
	unsigned long interrupt_state;
	int result;

	int chars = 0, size = 0, read = 0;
	char *pipebuf;

	if (filp->f_flags & O_NONBLOCK) {
		/*      if (RTF_LOCK(minor))
		   return -EAGAIN;
		 */
		if (RTF_EMPTY(minor))
			return 0;
	} else
		while (RTF_EMPTY(minor)  ) {
			if (check_blocked(current->signal, current->blocked) )
				return -ERESTARTSYS;
			RTF_SLEEPS(minor) = current;
			interruptible_sleep_on(&RTF_WAIT(minor));
		}
/*      RTF_LOCK(minor)++; */
	while (count > 0 && (size = RTF_LEN(minor))) {
		chars = RTF_MAX_RCHUNK(minor);
		if (chars > count)
			chars = count;
		if (chars > size)
			chars = size;

		read += chars;
		pipebuf = RTF_BASE(minor) + RTF_START(minor);
		count -= chars;
		copy_to_user(buf, pipebuf, chars); 
		rt_fifo_critical(interrupt_state);
		RTF_START(minor) += chars;
		RTF_START(minor) = RTF_WRAP(minor, RTF_START(minor));
		RTF_LEN(minor) -= chars;
		rt_end_fifo_critical(interrupt_state);
		buf += chars;
	}
/*      RTF_LOCK(minor)--; */
	if (read) {
		inode->i_atime = CURRENT_TIME;
		if (RTF_HANDLER(minor)) {
			if ((result = (*RTF_HANDLER(minor)) (minor)) < 0) {
				return result;
			}
		}
		return read;
	}
	return 0;
}


static ssize_t rtf_write(struct file *filp, const char *buf, size_t count, loff_t* ppos)
{
	struct inode* inode = filp->f_dentry->d_inode;
	unsigned int minor = MINOR(inode->i_rdev);
	unsigned long interrupt_state;
	int chars = 0, free = 0, written = 0;
	char *pipebuf;
	int result;

	if (count <= RTF_BUF(minor))
		free = count;
	else
		free = 1;

	while (count > 0) {
		while (RTF_FREE(minor) < free)  {
			if (check_blocked(current->signal, current->blocked) )
				return written ? : -ERESTARTSYS;
			if (filp->f_flags & O_NONBLOCK)
				return written ? : -EAGAIN;
			RTF_SLEEPS(minor) = current;
			interruptible_sleep_on(&RTF_WAIT(minor));
		}
		/*      RTF_LOCK(minor)++; */
		while (count > 0 && (free = RTF_FREE(minor))) {
			chars = RTF_MAX_WCHUNK(minor);
			if (chars > count)
				chars = count;
			if (chars > free)
				chars = free;
			rt_fifo_critical(interrupt_state);
			pipebuf = RTF_BASE(minor) + RTF_END(minor);
			rt_end_fifo_critical(interrupt_state);
			count -= chars;
			written += chars;

			copy_from_user(pipebuf, buf, chars);
			rt_fifo_critical(interrupt_state);
			RTF_LEN(minor) += chars;
			rt_end_fifo_critical(interrupt_state);
			buf += chars;
		}
		/*      RTF_LOCK(minor)--; */
		free = 1;
	}
	inode->i_ctime = inode->i_mtime = CURRENT_TIME;
	if (RTF_HANDLER(minor)) {
		if ((result = (*RTF_HANDLER(minor)) (minor)) < 0) {
			return result;
		}
	}
	return written;
}

static	unsigned int rtf_poll(struct file *filp, poll_table *wait)
{
	struct inode* inode = filp->f_dentry->d_inode;
	unsigned int minor = MINOR(inode->i_rdev);
	RTF_SLEEPS(minor) = current;
	poll_wait(filp, &RTF_WAIT(minor), wait);
	if ( ! RTF_EMPTY(minor)) {
		return POLLIN | POLLRDNORM;
	} else {
		return POLLOUT | POLLWRNORM;
	}
	return 0;
}


static struct file_operations rtf_fops =
{
	rtf_llseek,
	rtf_read,
	rtf_write,
	NULL,
	rtf_poll,
	NULL,
	NULL,
	rtf_open,
	NULL,
	rtf_release,
	NULL,
	NULL,
	NULL,
	NULL,
	NULL
};



int rtf_init(void)
{
	int i;

	if (register_chrdev(RTF_MAJOR, "rtf", &rtf_fops)) {
		printk("RT-FIFO: unable to get major %d.\n", RTF_MAJOR);
		return -EIO;
	}
	for (i = 0; i < RTF_NO; i++) {
		RTF_OPEN(i) = 0;
	}

	return 0;
}


#ifdef MODULE

int init_module(void)
{
	if ((sysrq.srq = rtf_init()) < 0) {
		return sysrq.srq;
	}
	if ((sysrq.srq = rt_request_srq(rt_fifo_sysrq_handler)) < 0) {
		printk("RT-FIFO: no sysrq available.\n");
		return sysrq.srq;
	}
	sysrq.in  = 0;
	sysrq.out = 0;
      	spin_lock_init(&sysrq.spinlock);
	return 0;
}


void cleanup_module(void)
{
	if (rt_free_srq(sysrq.srq) < 0) {
		printk("RT-FIFO: srq %d illegal or already free.\n", sysrq.srq);
	}
	unregister_chrdev(RTF_MAJOR, "rtf");
}

#endif				/* MODULE */
