#ifndef __RTF__
#define __RTF__

#ifdef __KERNEL__


#define RTF_MAJOR 63


extern int rtf_init(void);

/* attach a handler to an RT-FIFO
 *
 * Allow function \arg{handler} to be called when a user process reads or
 * writes to the FIFO. When the function is called, it is passed the fifo
 * number as the argument.
 */

extern int rtf_create_handler(unsigned int fifo,	/* RT-FIFO */
		int (*handler)(unsigned int fifo)	/* function to be called */);



/* create an RT-FIFO
 * 
 * An RT-FIFO \arg{fifo} is created with initial size of \arg{size}.
 * 
 * Return value: On success,  \arg{size} is returned. On error, -1 is returned.
 */

extern int rtf_create(unsigned int fifo, int size);

/* reset an RT-FIFO
 * 
 * An RT-FIFO \arg{fifo} is reset by setting its buffer pointers to zero, so
 *
 * that any iexisting data is discarded and the fifo started anew like at its
 *
 * creation.
 * 
 */

extern void rtf_reset(unsigned int fifo);


/* create an RT-FIFO using a bottom half queue
 * 
 * An RT-FIFO \arg{fifo} is created with initial size of \arg{size}.
 * 
 * The fifo uses bh_list to queue waiting Linux processes. If zero an immediate
 *
 * wake up is adopted. Create above defaults to tq_immediate. 
 * 
 * Return value: On success,  \arg{size} is returned. On error, -1 is returned.
 */

extern int rtf_create_using_bh(unsigned int fifo, int size, task_queue *bh_list);

/* create an RT-FIFO using a bottom half queue and a buffer assigned by the
 * user.
 */
extern int rtf_create_using_bh_and_usr_buf(unsigned int fifo, char *buf, int size, task_queue *bh_list);



/* destroy an RT-FIFO
 * 
 * Return value: On success, 0 is returned
 */

extern int rtf_destroy(unsigned int fifo);

extern int rtf_destroy_using_usr_buf(unsigned int fifo);



/* resize an RT-FIFO
 * 
 * Return value: \arg{size} is returned on success. On error, a negative value
 * is returned
 * 
 * Bugs: This operation destroys the contents of the FIFO
 */

extern int rtf_resize(unsigned int minor, int size);



/*
 * write to an RT-FIFO
 *
 * Try to write \arg{count} bytes to an FIFO.  Return -1 if there's not enough space
 * in the FIFO; otherwise return \arg{count}
 */

extern int rtf_put(unsigned int fifo,	/* RT-FIFO */
		void * buf,		/* buffer address */
	       	int count		/* number of bytes to write */);



/*
 * read from an RT-FIFO
 *
 * Try to read \arg{count} bytes from a FIFO.
 * Return -1 if there's not enough data in the FIFO; 
 * otherwise return \arg{count}
 */
extern int rtf_get(unsigned int fifo,	/* RT-FIFO */
		void * buf, 		/* buffer address */
		int count		/* number of bytes to read */);

#endif

#endif
